import sys
import numpy as np
from numpy import array
import matplotlib.pyplot as plt
MAX_GENERALIZATION = 20

def evaluate(z):
    categories = {}
    count_attr = [0]*MAX_GENERALIZATION #count number of records passed per level of generalization
    count_cat = {} #count distinct categories/attributes passed per level of generalization
    j = 0
    for line in open("simulation_output.txt","r"):
        items = line.split("\t")
        for i,count in enumerate(items[3::][::-1]):         
            if (int(count) >= z):
                if (i == 0):
                    lev = items[2]
                else: 
                    lev = '*'.join(items[2].split('*')[:-i])
                j += 1
                count_attr[i] += 1              
                if i not in count_cat:
                    count_cat[i] = [lev]
                else: 
                    if lev not in count_cat[i]:
                        count_cat[i].append(lev)                
                if(lev not in categories):
                    categories[lev] = 1
                else: categories[lev] += 1 
                break
    dict = {k: v for k, v in sorted(categories.items(), key=lambda item: item[1], reverse = True)}
    f = open("categories_zanon.txt", "w")  #store top categories 
    for i in dict:
        f.write(i + ": " + str(dict[i]) + "\n")
    f.close()
    
    fig = plt.figure(figsize=(8,8))
    plt.barh(list(dict.keys())[:40],list(dict.values())[:40])
    plt.gca().invert_yaxis() 
    plt.title("Top 40 attributes/categories over threshold with z = " + str(z))
    plt.tight_layout()
    plt.savefig('categories.pdf')
    
    print("Outputs over threshold with z = " + str(z) + ": " + str(j))
    for index,i in enumerate(count_attr):
        if((i) == 0):
            break
        print("With " + str(index) + "-generalization: " + str(i) + ". Distinct: " + str(len(count_cat[index])))
    print("Number of distinct outputs over threshold: " + str(len(categories)))
    
def evaluate_cat(zeta):
    evaluate(zeta)
    


import matplotlib.pyplot as plt
import numpy as np
import json
from collections import defaultdict
import pathlib

def plot_ca():
    params = json.loads(open('output_params.json', 'r').read())
    start = params['start']
    c_oa = params['c_oa']
    t_oa = params['t_oa']
    observed_attributes = params['observed_attributes']
    plt.figure(figsize = (10,6))
    plt.xlabel('s')
    plt.ylabel('c_a')
    #plt.plot([0, stop-start], [z, z], '--', label = 'z')
    for oa in observed_attributes:
        plt.plot([(x - start) for x in t_oa[oa]], c_oa[oa], label = oa)

    #plt.xticks()
    plt.grid()
    plt.legend()
    plt.show()

def get_pkanon_with_cat(output, z):
    pathlib.Path('./output/').mkdir(parents=True, exist_ok=True) 
    f = open('output/out_{:02d}.txt'.format(z), 'w+')
    attribute_set = set()
    final_dataset = defaultdict(set)
    for o in output:
        record = o[0] #tuple in the form (timestamp, user, attribute)
        c_a = o[1] #list of counters per category, most general to the left
        for i, level in enumerate(reversed(c_a)): #read the counters from the one reffering to the most specific category
            level = int(level)
            if level >= z: #if the counter satisfy the threshold, select the appropriate level of generalization
                if i == 0:
                    a = record[2] #if the most specific category satisfies the threshold, output the whole attribute
                else:
                    a = '*'.join(record[2].split('*')[:-i]) #else, remove the z-private specifications
                final_dataset[record[1]].add(a)
                if a not in attribute_set:
                    attribute_set.add(a)
                    f.write(a + '\n')
                break
    f.close()
    final_dataset_inv = defaultdict(list)
    for k,v in final_dataset.items():
            final_dataset_inv[str(v)].append(k)

    #ks = np.array([len(v) for v in final_dataset_inv.values()])
    return final_dataset_inv

def get_pkanon_without_cat(output, z):
    final_dataset = defaultdict(set)
    for o in output:
        record = o[0]
        c_a = int(o[1][-1])
        if c_a >= z:
            final_dataset[record[1]].add(record[2])

    final_dataset_inv = defaultdict(list)
    for k,v in final_dataset.items():
            final_dataset_inv[str(v)].append(k)

    #ks = np.array([len(v) for v in final_dataset_inv.values()])
    return final_dataset_inv

        
def pkanon_vs_z():
    output = []
    for line in open('simulation_output.txt', 'r'):
        items = line.split('\t')
        output.append(((float(items[0]), items[1], items[2]), [x for x in items[3:]]))
    
    z_range = range(1, 51)
    #k1 = []
    k2 = []
    k3 = []
    k4 = []
    k2_nocat = []
    for z in z_range:
        ks = np.array([len(v) for v in get_pkanon_with_cat(output, z).values()])
        k2.append(sum(ks[ks >= 2])/sum(ks))
        k3.append(sum(ks[ks >= 3])/sum(ks))
        k4.append(sum(ks[ks >= 4])/sum(ks))
        ks_nocat = np.array([len(v) for v in get_pkanon_without_cat(output, z).values()])
        if (sum(ks_nocat) == 0):
            break
        k2_nocat.append(sum(ks_nocat[ks_nocat >= 2])/sum(ks_nocat))
    
    plt.figure()
    plt.xlabel('z')
    plt.ylabel('p_kanon')
    plt.plot(z_range, k2, label = 'w/ categories')
    plt.plot(z_range, k2_nocat, label = 'w/o categories')
    plt.ylim(bottom = 0., top = 1.)
    plt.legend()
    plt.savefig('pkanon_cat_nocat.pdf')
    plt.show()
    
    plt.figure()
    plt.xlabel('z')
    plt.ylabel('p_kanon')
    plt.plot(z_range, k2, label = 'k = 2')
    plt.plot(z_range, k3, label = 'k = 3')
    plt.plot(z_range, k4, label = 'k = 4')
    plt.ylim(bottom = 0., top = 1.)
    plt.legend()
    plt.savefig('pkanon_vs_k.pdf')
    plt.show()
    
def evaluate_output():
    #plot_ca()
    pkanon_vs_z()
    
    
    


from utils import *
from evaluate_category import *
from evaluate_output import *
import matplotlib.pyplot as plt
import collections
import json


class zanon(object):

    def __init__(self, deltat, pk, k): 
        super(zanon, self).__init__()
        self.deltat = deltat
        self.rate = 1800
        self.z = 10
        self.H = {}
        self.c = {}
        self.pk = pk
        self.k = k
        self.t_start = 0
        self.t_stop = 0
        self.last_update = 0
        self.test = []
        self.time = []
        self.kanon = []
        self.queue = collections.deque()
        self.all_tuple = collections.deque()
        self.f_out = open('output.txt', 'w+')
        self.f_count = open('counters.txt', 'w+')
        self.tot_data = []
        self.details = [0]*20
        self.all_details = [[] for i in range(20)]
        self.anonymized  = collections.deque()
        self.tot_anon = []
        self.from_ll_to_mt = Transformer.from_crs('epsg:4326', 'epsg:3857')
        self.stepsize = np.array([3000, 5000, 10000, 30000, 500000])
        self.numeric_category = False

	
    def anonymize(self, tupla):
        
        t = float(tupla[0])
        u = tupla[1]  
       
        if(len(tupla) == 4):
            lat = tupla[2]
            lon = tupla[3]
            a = "*".join(str(x) for x in reversed(get_cell(self,lat,lon)))
            self.numeric_category = True
        elif(len(tupla)==3):
            a = tupla[2].strip() 
        else: raise ValueError("Arguments can be either 3 or 4, not " + str(len(tupla)))
            
              
        if self.t_start == 0:
            self.t_start = t
            output = {}
            output['all_details']=[[] for i in range(20)]
            output['tot_anon']=[]
            output['time'] = []
            output['z']=[]
            output['kanon']=[]
            with open("output.json","w+") as f:
                json.dump(output,f)
             
        sep = '*'
        cat = a.split(sep)
        for level in range(len(cat)):    
            att = '*'.join(cat[:level + 1])
            if att in self.H:
                evict(self, t, att)
                  
        clean_queue(self,t)        
        z_change(self, t)          
        self.t_stop = t                
        manage_data_structure(self, t, u, a)       
        check_and_output(self, t, u, a)
               
    def duration(self):
        print('End of simulation (simulated time: {})'.format(str(timedelta(seconds = int(self.t_stop - self.t_start)))))

    def evaluate_output(self):
        evaluate_output()
        
    def evaluate_category(self,z):
        evaluate_cat(z)
        
    def final_kanon(self):
        final_kanon()

    def plot_z(self):
        plot_z(self)
        
    def endFiles(self):
        self.f_out.close()
        self.f_count.close()
		


import numpy as np
from collections import defaultdict
import subprocess
import sys
import collections
from datetime import datetime, timedelta
from pyproj import Transformer
import json


def get_cell(self, lat, lon):
    
    nw = self.from_ll_to_mt.transform(47.5, 5.5)
    se = self.from_ll_to_mt.transform(35, 20)
    point = self.from_ll_to_mt.transform(lat, lon)
    distx = point[0] - nw[0]
    disty = nw[1] - point[1]
    cellx = (distx / self.stepsize).astype(int)
    celly = (disty / self.stepsize).astype(int)
    maxx = (np.ceil((se[0] - nw[0]) / self.stepsize)).astype(int)
    cell = celly * maxx + cellx
    return cell.tolist()


def kanon_for_binary(self, z):
    final_dataset = defaultdict(set)
    for i in self.all_tuple:      #coda di tuple(t,u,a,contatori di a)     
        u = i[1]
        a = i[2]
        counters = i[3]
        cat = a.split("*")
        att = None
        for i,c in enumerate(counters):
            if int(c) >= z:
                att = '*'.join(cat[:i + 1]) 
        if att != None:
            final_dataset[u].add(att) 
            
    if len(final_dataset) != 0:  
        final_dataset_inv = defaultdict(list)
        for k,v in final_dataset.items():
            final_dataset_inv[str(v)].append(k)    
        ks = np.array([len(v) for v in final_dataset_inv.values()])
        #print(sum(ks[ks >= 2])/sum(ks))
        return sum(ks[ks >= self.k])/sum(ks)
    else: return 2
                

def binary_search(self, pk, z_start, z_end): 
    if z_start > z_end:
        return -1

    z_mid = (z_start + z_end) // 2 
    r = kanon_for_binary(self, int(z_mid))
    if r >= pk and r <= pk + 0.01:
        return z_mid

    if r > pk:
        return binary_search(self, pk, z_start, z_mid-1)
    else:
        return binary_search(self, pk, z_mid+1, z_end)
    


def z_change(self, t):
    if(t - self.t_start >= self.deltat and t-self.last_update >= self.rate):
        self.last_update = t
        result = binary_search(self, self.pk, 0, 1800)
        '''
        self.kanon.append(result)
        if(result < 0.80):
            self.z += 1
        if(result > 0.81 and self.z > 5):
            self.z -=1
        #print("Value of z: " + str(self.z))
        '''
        with open("output.json", "r") as fi:
            data = json.load(fi)
            if(result != -1):
                self.z = result
            i = 0
            for  detail in self.details:
                #self.all_details[i].append(detail)
                data['all_details'][i].append(detail)
                i += 1
            data['tot_anon'].append(len(self.anonymized))
            data['time'].append(str(datetime.utcfromtimestamp(t).strftime('%Y-%m-%d %H:%M:%S')))
            data['z'].append(self.z)
            data['kanon'].append(compute_kanon(self))
        with open("output.json", "w") as fi:
            json.dump(data, fi)

def compute_kanon(self):   
    
    final_dataset = defaultdict(set)
    for i in self.queue:
        final_dataset[i[1]].add(i[2])       
    final_dataset_inv = defaultdict(list)
    for k,v in final_dataset.items():
        final_dataset_inv[str(v)].append(k)
    
    ks = np.array([len(v) for v in final_dataset_inv.values()])
    if(sum(ks) != 0):
        return sum(ks[ks >= 2])/sum(ks)
    else: return 1.0
    

def clean_queue(self,t):
    
    if self.all_tuple:
        while True:
            if self.all_tuple:
                temp = self.all_tuple.popleft()          
                if(t - temp[0] <= self.deltat):
                    self.all_tuple.appendleft(temp)
                    break
            else: break
    
    if self.queue:            
        while True:
            if self.queue:
                temp = self.queue.popleft()     
                detail = len(temp[2].split("*"))
                self.details[detail - 1] -= 1
                if(t - temp[0] <= self.deltat):
                    self.queue.appendleft(temp)
                    self.details[detail - 1] += 1
                    break
            else: break
         
    if self.anonymized:            
        while True:
            if self.anonymized:
                temp = self.anonymized.popleft()
                if(t - temp[0] <= self.deltat):
                    self.anonymized.appendleft(temp)
                    break
            else: break
    

def read_next_visit(line):
    t, u, a = line.split(',')
    t = float(t)
    a = a.strip()
    return t, u, a
    
def a_not_present(self, t, u, a):
    self.H[a] = collections.OrderedDict()
    self.H[a][u] = t
    self.c[a] = 1
    


def a_present(self, t, u, a):
    if u not in self.H[a]:
        u_not_present(self, t, u, a)
    else:
        u_present(self, t, u, a)

def u_not_present(self, t, u, a):
    self.H[a][u] = t
    self.c[a] += 1

    
def u_present(self, t, u, a):
    self.H[a][u] = t
    self.H[a].move_to_end(u)
    
def evict(self, t, a):
    to_remove = []
    for u,time in self.H[a].items():
        if (t - time > self.deltat):
            to_remove.append(u)
        else:break
    for u in to_remove:
        self.H[a].pop(u, None)
        self.c[a] -= 1
        if len(self.H[a]) == 0:          
            self.H.pop(a, None)
            break


def manage_data_structure(self, t, u, a):
    sep = '*'
    cat = a.split(sep)
    for level in range(len(cat)):
        i = '*'.join(cat[:level + 1])
        if i not in self.H:
            a_not_present(self, t, u, i)
        else:
            a_present(self, t, u, i)
            

        
def check_and_output(self, t, u, a):
    sep = '*'
    cat = a.split(sep)
    counters = []
    output = None
    for level in range(len(cat)):
        attr = '*'.join(cat[:level + 1])
        counters.append(self.c[attr])
        if self.c[attr] >= self.z:
            output = (t,u,attr) 
            
    if(output != None):
        self.queue.append(output)
        detail = len(output[2].split("*"))
        self.details[detail - 1] += 1
        self.f_out.write("\t".join(str(x) for x in output) + "\n")
    else: self.anonymized.append((t,u,a))
    self.f_count.write("\t".join(str(x) for x in [t,u,a])+ "\t" +
                       "\t".join(str(x) for x in counters) + '\n')
    self.all_tuple.append((t,u,a,counters))
    



from collections import defaultdict
import numpy as np
import json
import matplotlib.pyplot as plt
from matplotlib.ticker import LinearLocator

MAX_GENERALIZATION = 20
stepsize = np.array([3000, 5000, 10000, 30000, 500000])

numeric_category = False

with open('output.json',"r") as f:
    data = json.load(f)
    labels = []   
    labels.append("Anonymized")
    details = [x for x in data["all_details"] if not all(v == 0 for v in x)]
    if numeric_category:
        for i in reversed(stepsize):
            labels.append(str(int(i/1000))+" km")
    else:
        for i in range(len(details)):
            labels.append(str(i + 1) + "-detail")       
    to_plot =  np.vstack((data["tot_anon"], details))
    color = 'tab:red'
    fig, ax_left = plt.subplots(figsize=(20, 10))
    ax_right = ax_left.twinx()
    ax_third = ax_left.twinx()
    ax_third.spines["right"].set_position(("axes", 1.1))
    ax_left.plot(data["time"],data["z"], color=color, linewidth="5")
    ax_right.plot(data["time"], data["kanon"], color='black', linewidth="5")
    ax_third.stackplot(data["time"], to_plot, 
                        labels = labels ,alpha=0.4)
    ax_third.legend(loc='upper left', prop={"size":20})
    ax_left.set_xlabel('time', fontsize=20)
    ax_left.set_ylabel('z', color=color, fontsize=20)
    ax_left.autoscale()
    ax_third.autoscale()
    ax_third.set_ylabel('Tuple traffic', color = "blue", fontsize=20)
    ax_third.tick_params(axis='y', labelcolor="blue", labelsize=20.0)
    ax_right.set_ylim(bottom = 0.0, top = 1.0)
    ax_left.tick_params(axis='y', labelcolor=color, labelsize=20.0)
    ax_right.set_ylabel('pkanon', color='black', fontsize= 20)
    ax_right.tick_params(axis='y', labelcolor='black', labelsize = 20.0)
    ax_left.get_xaxis().set_major_locator(LinearLocator(numticks=20))
    ax_left.tick_params(labelsize=20)        
    fig.autofmt_xdate(rotation = 45)
    fig.tight_layout()
    fig.savefig('z_tuning.pdf')

with open('trace.txt') as f:
    rows =  sum(1 for _ in f)
    
final_dataset = defaultdict(set)
file = open('output.txt','r')
gen = [0]*MAX_GENERALIZATION
tot = 0
for line in file:
    tot += 1
    t,u,a = line.split("\t")
    t = float(t)
    a.strip()          
    final_dataset[u].add(a)
    cat = a.split("*")
    gen[len(cat)] += 1

final_dataset_inv = defaultdict(list)
for k,v in final_dataset.items():
    final_dataset_inv[str(v)].append(k)
ks = np.array([len(v) for v in final_dataset_inv.values()])
for k in range(2,5):
    print("Final " + str(k) + "-anonymization: " + str(sum(ks[ks >= k])/sum(ks)))
    
for index,i in enumerate(gen):
    if(i == 0 and index == 0):
        continue
    elif(i == 0):
        break
    print("Tuple passed with " + str(index ) + "-details level: " + str(i))
print("Tuple anonymized: " + str(rows - tot))
    


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="zanon", 
    version="0.3.2",
    author="SmartData@PoliTO",
    author_email="s251325@studenti.polito.it",
    description="A module for z-anonymity",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/pimcity/wp2/zeta-anonymity",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)